#!/usr/bin/python

import os
import random
import sys
import time

import termios
import tty

import handlers
import negotiators


# --------------------------------------------------------
#   static
# --------------------------------------------------------
KEY_ESC = 27
KEY_ENTER = 13
KEY_SPACE = 32

SPACE_BUFFER = ' '*5
NSPACE_BUFFER = '\n%s'%SPACE_BUFFER

COUNT_STATUS_LINES_RESERVED = 5

ROWSIZE = 18
COLSIZE = 50


# --------------------------------------------------------
#   ascii escapes
# --------------------------------------------------------
#
# http://ascii-table.com/ansi-escape-sequences.php
BG_COLOURS = { 'black': 40
             , 'red': 41
             , 'green': 42
             , 'yellow': 43
             , 'blue': 44
             , 'magenta': 45
             , 'cyan': 46
             , 'white': 47
             }

FG_COLOURS = { 'black': 30
             , 'red': 31
             , 'green': 32
             , 'yellow': 33
             , 'blue': 34
             , 'magenta': 35
             , 'cyan': 36
             , 'white': 37
             }

ENDC = '\033[0m'

def cformat(string, fg, bg='black', bold=False):
    if bold:
        ibold = 1
    else:
        ibold = 3
    return '\033[%s;%s;%sm%s%s'%( FG_COLOURS[fg]
                                , BG_COLOURS[bg]
                                , ibold
                                , string
                                , ENDC
                                )


# --------------------------------------------------------
#   entity statics
# --------------------------------------------------------
ENTITY_SIZE_MULTIPLE = 5

ES_PILL = 1
ES_RING = ES_PILL * ENTITY_SIZE_MULTIPLE
ES_VIAL = ES_RING * ENTITY_SIZE_MULTIPLE
ES_TOOL = ES_VIAL * ENTITY_SIZE_MULTIPLE
ES_TOME = ES_TOOL * ENTITY_SIZE_MULTIPLE
ES_STOOL = ES_TOME * ENTITY_SIZE_MULTIPLE
ES_PLINTH = ES_STOOL * ENTITY_SIZE_MULTIPLE
ES_WALL = ES_PLINTH * ENTITY_SIZE_MULTIPLE


# --------------------------------------------------------
#   tile statics
# --------------------------------------------------------
SPACE_CAPACITY = ES_WALL


# --------------------------------------------------------
#   cardinals
# --------------------------------------------------------
C_SW = '1'
C_SS = '2'
C_SE = '3'
C_WW = '4'
C_EE = '6'
C_NW = '7'
C_NN = '8'
C_NE = '9'


# --------------------------------------------------------
#   container
# --------------------------------------------------------
#
# A container holds entities. e.g. a player's inventory. The
# base case looks very much like a list, but containers need
# logic for things like capacity checking and placement.
#
# Entities always reside in something that quacks like a
# container.
#
class Container(object):
    def __init__(self, context, owner):
        "owner is not necessaily an ent."
        self.context = context
        self.owner = owner

        self._within = []
    def __len__(self):
        return len(self._within)
    def __iter__(self):
        for ent in self._within:
            yield ent
    def things(self):
        sb = []
        for itm in self._within[:-1]:
            if itm.mobile:
                sb.append('%s %s (*)'%(str(itm), itm.name))
            else:
                sb.append('%s %s'%(str(itm), itm.name))
        return ', '.join(sb)
    def contains(self, ent):
        return ent in self._within
    def adopt(self, ent):
        if ent.contained_by and ent in ent.contained_by._within:
            ent.contained_by._within.remove(ent)
        ent.contained_by = self
        self._within.append(ent)
        # keep track of the top tile claimers back on top
        claimers = []
        for went in self._within:
            if went.claims_tile:
                claimers.append(went)
        for claimer in claimers:
            self._within.remove(claimer)
            self._within.append(claimer)
    def abandon(self, ent):
        ent.container_by = None
        self._within.remove(ent)
    def remove(self, ent):
        self._within.remove(ent)
        ent.contained_by = None
    def top(self):
        return self._within[-1]
    def top_under_player(self):
        return self._within[-2]
    def empty(self):
        self._within = []


# --------------------------------------------------------
#   context struct
# --------------------------------------------------------
#
# This gives us a nametile that we can pass around the
# program. It's a centralised place for storing game
# state, and a the place to start when you want to reach
# an object known to the game.
#
class Context(object):
    def __init__(self):
        self.keep_running = True
        self.exit_message = None
        self.clock = 0

        self.io = None

        self.status = []

        # int v object
        self.blocks = {}
        self.unique_block_idx = -1

        # fairies
        self.fairies = []
        self.player = None

        self.starting_block = None

        # (tool, ground) v CraftRule
        self.craftbook = None
        # name v fn_that_declares_an_entity
        self.entity_stencils = None
        # (actor, target) v negotiator_fn
        self.negotiators = None
    def retain_block(self, block):
        self.unique_block_idx += 1
        block.unique_id = self.unique_block_idx
        self.blocks[self.unique_block_idx] = block
    def release_block(self, block_id):
        del self.blocks[block_id]


# --------------------------------------------------------
#   io
# --------------------------------------------------------
class UnixIo(object):
    def __init__(self, context):
        self.context = context
    def clear_screen(self):
        os.system('clear')
        #print chr(27) + "[2J"
    def getchar(self):
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch
    def ask_yn(self, msg, default, clear=True):
        def say_no():
            print 'No'
            time.sleep(0.1)
        def say_yes():
            print 'Yes'
            time.sleep(0.1)

        if clear:
            self.clear_screen()

        print '\n%s'%msg,
        if None != default:
            if default:
                print "[y]"
            else:
                print "[n]"
        else:
            print
        r = self.getchar()
        if r in ['y', 'Y']:
            say_yes()
            return True
        if r in ['n', 'N']:
            say_no()
            return False

        if default:
            say_yes()
            return True
        else:
            say_no()
            return False
    def get_wordy_input(self):
        sys.stdout.write(':')
        input = raw_input()
        return input
    def redraw(self, lines, with_status=False):
        status_buffer = COUNT_STATUS_LINES_RESERVED - len(self.context.status)

        sb = []
        #
        # status messages, or gaps in their stead
        for i in range(status_buffer):
            sb.append('')
        if with_status:
            sb.extend(self.context.status)
        #
        # render the supplied lines
        sb.extend(lines)

        out = "%s%s"%(SPACE_BUFFER, NSPACE_BUFFER.join(sb))

        self.clear_screen()
        print out


# --------------------------------------------------------
#   mapping
# --------------------------------------------------------
#
# Blocks are composed of Tile. Blocks can have neighbours.
# The neighbours don't have to be known at start time, as
# they become relevant you can have a Biome generate
# them.
#
class Block(object):
    def __init__(self, context, biome, rowsize, colsize):
        self.unique_id = None

        self.context = context
        self.biome = biome
        self.rowsize = rowsize
        self.colsize = colsize

        self.d_tile = {}
        # numbers represent keypad cardinals
        self.neighbours = { C_SE: None
                          , C_SS: None
                          , C_SW: None
                          , C_WW: None
                          , C_EE: None
                          , C_NW: None
                          , C_NN: None
                          , C_NE: None
                          }
    def __getitem__(self, coords):
        return self.d_tile[coords]
    def retain_entity(self, row, col, placement):
        coords = (row, col)
        self.placements[coords] = placement
    def release_entity(self, entity):
        ridx, cidx = [(r, c) for (r, c), p in self.placements.items() if p == entity][0]
        coords = (ridx, cidx)
        del self.placements[coords]
    def render(self):
        rows = []
        for ridx in xrange(self.rowsize):
            sb = []
            for cidx in xrange(self.colsize):
                sb.append(self.d_tile[ (ridx, cidx) ].symbol())
            rows.append(sb)
        return [''.join(colchars) for colchars in rows]
    def move(self, ent, keypad_cardinal):
        #
        # Work out the desired coordinate
        block_id, old_ridx, old_cidx = ent.contained_by.owner.get_position()
        block = self.context.blocks[block_id]
        nrow, ncol = old_ridx, old_cidx
        if keypad_cardinal in (1,2,3):
            nrow += 1
        if keypad_cardinal in (1,4,7):
            ncol -= 1
        if keypad_cardinal in (3,6,9):
            ncol += 1
        if keypad_cardinal in (7,8,9):
            nrow -= 1
        #
        # Validate it's within map bounds
        if nrow < 0 or nrow > self.rowsize-1:
            return
        elif ncol < 0 or ncol > self.colsize-1:
            return
        new_tile = self.d_tile[(nrow, ncol)]
        #
        # If the user has motioned towards claimed tile, that's
        # an interaction rather than movement.
        target = new_tile.claimant()
        if target:
            key = (ent.name, target.name)
            if key in self.context.negotiators:
                fn_name = self.context.negotiators[key]
                fn = negotiators.__getattribute__(fn_name)
                fn(self.context, ent, target)
            else:
                status(self.context, '*bump*')
            return
        #
        # Do the move
        new_tile.container.adopt(ent)
        ent.facing = keypad_cardinal

def create_block(context, rowsize, colsize):
    block = Block(context, None, rowsize, colsize)
    context.retain_block(block)
    #
    # Prepare the tile containers in the block
    for ridx in xrange(block.rowsize):
        for cidx in xrange(block.colsize):
            tile = create_tile(block.context, block, ridx, cidx)
            block.d_tile[ (ridx, cidx) ] = tile

    return block

def merge_block_onto_block(context, source_block, target_block, t_ridx_offset, t_cidx_offset):
    """Merges all of source onto target, using (t_ridx_offset, t_cidx_offset)
    as the top-left point at which to start merging."""
    for ridx in xrange(source_block.rowsize):
        for cidx in xrange(source_block.colsize):
            tile = source_block.d_tile[ (ridx, cidx) ]
            tile.block = target_block
            tile.ridx = ridx + t_ridx_offset
            tile.cidx = cidx + t_cidx_offset
            t_coords = ( tile.ridx
                       , tile.cidx
                       )
            target_block.d_tile[t_coords] = tile

class Tile(object):
    def __init__(self, context, block, ridx, cidx):
        self.context = context
        self.capacity = SPACE_CAPACITY
        self.container = Container(context, self)
        self.block = block
        self.ridx = ridx
        self.cidx = cidx
    def __sizeof__(self):
        return self.within[:]
    def symbol(self):
        return str(self.container.top())
    def claimant(self):
        chosen = None
        for ent in self.container:
            if ent.claims_tile:
                chosen = ent
        return chosen
    def get_position(self):
        """Returns a three-item tuple representing the
        global coordinate position of this container,
        (level, row, column)"""
        return (self.block.unique_id, self.ridx, self.cidx)

def create_tile(context, block, ridx, cidx):
    tile = Tile(context, block, ridx, cidx)
    hatch_entity(context, 'rough_sand', tile.container)
    return tile


# --------------------------------------------------------
#   sketch functions
# --------------------------------------------------------
#
# point to these from feature blocks
#
def sketch_bed(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'slab', tile_container)
    hatch_entity(context, 'bed', tile_container)
def sketch_dirt(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'dirt', tile_container)
def sketch_door(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'slab', tile_container)
    hatch_entity(context, 'door', tile_container)
def sketch_jug(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'slab', tile_container)
    hatch_entity(context, 'jug', tile_container)
def sketch_mat(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'slab', tile_container)
    hatch_entity(context, 'mat', tile_container)
def sketch_sand(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'sand', tile_container)
def sketch_outhouse_door(context, tile_container):
    tile_container.empty()
    od = hatch_entity(context, 'outhouse_door', tile_container)
    (
        hatch_entity(context, 'rake', od.inv)
    )
def sketch_slab(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'slab', tile_container)
def sketch_wall(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'dirt', tile_container)
    hatch_entity(context, 'wall', tile_container)
def sketch_well(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'dirt', tile_container)
    hatch_entity(context, 'well', tile_container)
def sketch_wood(context, tile_container):
    tile_container.empty()
    hatch_entity(context, 'dirt', tile_container)
    hatch_entity(context, 'wood', tile_container)


# --------------------------------------------------------
#   feature blocks
# --------------------------------------------------------
#
# You create structures on a block of their own, and then
# copy them onto the player's current block. You should
# either create the block once and store it in context so
# you can reuse it, or else be vigilant to deallocate the
# block once you're done with it! :)
#
def create_well_block(context):
    sketch = [ 'n_n'
             , '_w_'
             , 'n_n'
             ]
    block = create_block( context
                        , rowsize=len(sketch)
                        , colsize=len(sketch[0])
                        )
    legend = { 'n': sketch_wall
             , '#': sketch_wall
             , '_': sketch_slab
             , 'w': sketch_well
             }
    for ridx, row in enumerate(sketch):
        for cidx, cell in enumerate(row):
            setup_fn = legend[cell]
            if setup_fn:
                tile = block.d_tile[ (ridx, cidx) ]
                setup_fn(context, tile.container)

    return block

def create_house_block(context):
    sketch = [ '________'
             , '_/----n_'
             , '_|_12_|_'
             , '_+_==_+_'
             , '_|____|_'
             , '_n----/_'
             , '________'
             ]
    block = create_block( context
                        , rowsize=len(sketch)
                        , colsize=len(sketch[0])
                        )
    legend = { '/': sketch_wood
             , '|': sketch_wood
             , '-': sketch_wood
             , 'n': sketch_wood
             , '_': sketch_slab
             , 'w': sketch_well
             , '~': sketch_sand
             , '+': sketch_door
             , '=': sketch_bed
             , '1': sketch_mat
             , '2': sketch_jug
             }
    for ridx, row in enumerate(sketch):
        for cidx, cell in enumerate(row):
            setup_fn = legend[cell]
            if setup_fn:
                tile = block.d_tile[ (ridx, cidx) ]
                setup_fn(context, tile.container)

    return block

def create_outhouse_block(context):
    sketch = [ '|+|'
             , '|||'
             , '|||'
             ]
    block = create_block( context
                        , rowsize=len(sketch)
                        , colsize=len(sketch[0])
                        )
    legend = { '|': sketch_wood
             , '+': sketch_outhouse_door
             }
    for ridx, row in enumerate(sketch):
        for cidx, cell in enumerate(row):
            setup_fn = legend[cell]
            if setup_fn:
                tile = block.d_tile[ (ridx, cidx) ]
                setup_fn(context, tile.container)

    return block


# --------------------------------------------------------
#   entities
# --------------------------------------------------------
#
# An entity is anything that can have a sprite in the world.
#
class Entity(object):
    def __init__(self, context):
        self.context = context

        self.c = None
        self.name = None
        self.mass = None
        self.inv = None

        self.contained_by = None

        self.colour = None
        self.bold = None

        # sentient beings can act on things that claim a tile.
        self.claims_tile = False
        # can it be picked up?
        self.mobile = False
        #
        # item-specific properties and functions
        self.d = {}
        #
        # plugin functions
        self.desc_fn = None

        # callbacks for actions
        self.on_apply_fn = None
        self.on_wet_fn = None

        self.facing = C_SS
    def __repr__(self):
        return cformat(self.c, self.colour, bg=self.bg, bold=self.bold)
    def desc(self):
        """This is default, but you can override it by setting
        desc_fn."""
        if self.desc_fn:
            return self.desc_fn(self)
        else:
            return self.name

def hatch_entity(context, stencil_name, container):
    estencil = context.entity_stencils[stencil_name]

    ent = Entity(context)
    ent.c = estencil.symbol
    ent.name = estencil.name
    ent.colour = estencil.colour
    ent.bg = estencil.bg
    ent.bold = estencil.bold
    ent.claims_tile = estencil.claims
    ent.mobile = estencil.mobile

    if estencil.inv:
        ent.inv = Container(context, ent)
        ent.inv.owner = ent

    container.adopt(ent)

    return ent


# --------------------------------------------------------
#   entity stencils
# --------------------------------------------------------
ENTITY_DEFINITIONS = '''
ent ( bag
    c black
    bold
    mobile
    inv

ent * bean seed
    c green
    bold
    mobile

ent . bean plant (just planted)

ent j bean sapling
    c green
    claims

ent l bean plant
    c green
    claims

ent = bed
    c yellow
    claims

ent . dirt

ent + door
    claims

ent ( jug
    c blue
    mobile
    charges 5 0

ent ( mat
    c red
    mobile

ent + outhouse_door
    c cyan
    claims
    inv

ent @ player
    c yellow
    bold
    claims
    inv

ent ( rake
    mobile

ent _ raked_sand
    c yellow
    bold

ent ~ rough_sand
    c yellow

ent - slab

ent ( spade
    mobile

ent | wall
    claims

ent ~ well
    c blue
    bold
    claims

ent | wood
    c yellow
    claims
'''

class EntityStencil(object):
    """This is an object that describes how to construct
    a particular entity."""
    def __init__(self):
        self.name = None
        self.symbol = None
        self.colour = 'white'
        self.bg = 'black'
        self.bold = False
        self.claims = False
        self.mobile = False
        self.inv = False

        self.charges = True
        self.charge_max = None
        self.charge_now = None

def build_entity_stencils(context):
    context.entity_stencils = {}

    es = None
    for line in ENTITY_DEFINITIONS.split('\n'):
        if line.startswith('ent '):
            es = EntityStencil()
            lst_tpl = line.strip().split(' ')
            es.symbol = lst_tpl[1]
            es.name = ' '.join(lst_tpl[2:])

            if es.name in context.entity_stencils:
                m = 'duplicate definitions for [%s]'%es.name
                raise Exception(m)

            context.entity_stencils[es.name] = es
        else:
            l = line.strip().split(' ')
            if l[0] == '':
                continue
            elif l[0] == 'c':
                es.colour = l[1]
            elif l[0] == 'bold':
                es.bold = True
            elif l[0] == 'claims':
                es.claims = True
            elif l[0] == 'mobile':
                es.mobile = True
            elif l[0] == 'inv':
                es.inv = True
            elif l[0] == 'charges':
                es.charges = True
                es.charge_max = l[1]
                es.charge_now = l[2]
            else:
                raise Exception('unrecognised [%s]'%line.strip())


# --------------------------------------------------------
#   crafting
# --------------------------------------------------------
CRAFTING_DOCUMENT = """
#
# tool ground butterfly handlers message
rake/rough_sand/raked_sand    // Your strokes release a flow in the sand.
spade/rough_sand    //reheap
spade/raked_sand    //reheap
bag/dirt            //collect
"""

class CraftRule(object):
    def __init__(self):
        self.tool = None
        self.ground = None
        # the list of things that the object transforms into, if any.
        self.butterfly = []
        self.handlers = []
        self.message = None

def build_craftbook(context):
    context.craftbook = {}

    cd_lines_we_care_about = [l.strip() for l
                                        in CRAFTING_DOCUMENT.split('\n')
                                        if len(l.strip()) > 0
                                        ]
    # get rid of comments
    cd_lines_we_care_about = [s.split('#')[0] for s
                                              in cd_lines_we_care_about
                                              ]
    # parse the line into tuples
    cd_lines_we_care_about = [l.split('/') for l
                                           in cd_lines_we_care_about
                                           if len(l.strip()) > 0
                                           ]
    # convert it into tuples of right length
    def space_to_five(tpl):
        lst = list(tpl)
        while 5 > len(lst):
            lst.append(None)
        return lst
    rule_stencils = [space_to_five(line) for line
                                         in cd_lines_we_care_about
                                         ]

    named_entities = set()
    named_handlers = set()
    for rs in rule_stencils:
        rs_tool, rs_ground, rs_butterflies, rs_handlers, rs_msg = rs

        if ',' in rs_tool or ',' in rs_ground:
            m = "Bad rule. For tool and ground: one per line. [%s]"%str(rs)
            raise Exception(m)

        named_entities.add(rs_tool.strip())
        named_entities.add(rs_ground.strip())
        for b in [t.strip() for t in rs_butterflies.split(',') if len(t.strip())>0]:
            named_entities.add(b)
        for h in [t.strip() for t in rs_handlers.split(',') if len(t.strip())>0]:
            named_handlers.add(h)

        rule = CraftRule()
        rule.tool = rs_tool
        rule.ground = rs_ground
        rule.butterfly = [b.strip() for b
                                    in rs_butterflies.split(',')
                                    if len(b.strip()) > 0
                                    ]
        rule.handlers = [h.strip() for h
                                   in rs_handlers.split(',')
                                   if len(h.strip()) > 0
                                   ]
        if rs_msg:
            rule.message = rs_msg.strip()

        context.craftbook[ (rs_tool, rs_ground) ] = rule

    for e in named_entities:
        if e not in context.entity_stencils:
            raise Exception("No entity stencil for %s"%e)
    for h in named_handlers:
        if h not in dir(handlers):
            raise Exception("No handler for %s"%h)
    

# --------------------------------------------------------
#   negotiations
# --------------------------------------------------------
#
# A negotiation is what happens when an actor tries to move
# an avatar entity into a space that's already claimed by
# something else.
#
NEGOTIATION_RULES = '''
player/door/player_door
player/outhouse_door/player_outhouse_door
player/well/player_well
player/bed/player_sleep
'''

def build_negotiators(context):
    context.negotiators = {}

    nrules = [l.strip().split('/') for l
                                  in NEGOTIATION_RULES.split('\n')
                                  if len(l.strip()) > 0
                                  ]
    for tpl in nrules:
        if len(tpl) != 3:
            raise Exception("Invalid rule %s"%str(tpl))
        actor, target, fn_name = [t.strip() for t in tpl]
        if actor not in context.entity_stencils:
            raise Exception("Unknown entity %s"%actor)
        if target not in context.entity_stencils:
            raise Exception("Unknown entity %s"%target)
        if fn_name not in dir(negotiators):
            raise Exception("Unknown rule function %s"%fn_name)
        context.negotiators[ (actor, target) ] = fn_name


# --------------------------------------------------------
#   inventory
# --------------------------------------------------------
INVENTORY_SELECTION_KEYS = []
INVENTORY_SELECTION_KEYS.extend( [chr(c) for c in xrange(ord('a'), ord('z')+1)] )
INVENTORY_SELECTION_KEYS.extend( [chr(c) for c in xrange(ord('A'), ord('Z')+1)] )

def inspect_mobile_items_in_container(context, container, heading, use_selection):
    """Looks at a container. If selection is True, then you can select
    something from the container."""
    mobile_items = [tpl[1] for tpl in sorted([(e.desc(), e) for e
                                                            in container
                                                            if e.mobile
                                                            ])
                    ]

    choice_d = {}
    for idx, ent in enumerate(mobile_items):
        idesc = ent.desc()
        choice_d[INVENTORY_SELECTION_KEYS[idx]] = (idesc, ent)

    selection = None
    status = None

    while selection == None:
        lines = []
        lines.append(heading)
        lines.append('')
        for key in sorted(choice_d.keys()):
            name = choice_d[key][0]
            if use_selection:
                lines.append("  %s) %s"%(key, name))
            else:
                lines.append("  %s"%name)
        if len(choice_d) == 0:
            lines.append('  [Nothing]')
        lines.append('')
        if status:
            lines.append(status)

        context.io.redraw(lines)
        result = context.io.getchar()
        if not use_selection:
            break
        if ord(result) in (KEY_ESC, KEY_SPACE, KEY_ENTER):
            return None
        elif result in choice_d:
            selection = choice_d[result][1]
        else:
            status = "* escape to return"
    return selection


# --------------------------------------------------------
#   actions
# --------------------------------------------------------
KEYPRESS_ACTIONS = {}

def handle_keypress(*keys):
    def decorator(func):
        def wrapper(context):
            func(context)
        for key in keys:
            if key in KEYPRESS_ACTIONS:
                raise Exception("Duplicate keybind definition. %s"%key)
            KEYPRESS_ACTIONS[key] = wrapper
        if func.__doc__:
            wrapper.__doc__ = func.__doc__
        else:
            wrapper.__doc__ = func.__name__
        if wrapper.__doc__.startswith('action_'):
            wrapper.__doc__ = ' '.join(wrapper.__doc__.split('_')[1:])
        return wrapper
    return decorator

@handle_keypress('i', 's', '.')
def action_apply(context):
    ent = inspect_mobile_items_in_container( context=context
                                           , container=context.player.ent.inv
                                           , heading="Select item to apply"
                                           , use_selection=True
                                           )
    if None == ent:
        return

    space_container = context.player.ent.contained_by
    target = space_container.top_under_player()
    handler_key = (ent.name, target.name)
    if handler_key in context.craftbook:
        craft_rule = context.craftbook[handler_key]
        for fn_name in craft_rule.handlers:
            fn = handlers.__getattribute__(fn_name)
            fn(self.context, self.context.player.ent, ent, target)
        if craft_rule.message:
            status(context, craft_rule.message)
        if craft_rule.butterfly:
            space_container.abandon(target)
            for b in craft_rule.butterfly:
                hatch_entity(context, b, space_container)
        return
    else:
        status(context, 'Nothing to do')
        return

@handle_keypress('!', 'Q')
def action_quit(context):
    context.keep_running = False

@handle_keypress('?')
def action_help(context):
    "Help: shows this menu"
    context.io.clear_screen()
    sb = []
    for i in range(COUNT_STATUS_LINES_RESERVED):
        sb.append('')
    sb.append('%sKeys:'%SPACE_BUFFER)
    sb.append('')
    for k, fn in sorted(KEYPRESS_ACTIONS.items()):
        text = fn.__doc__
        sb.append('%s  %s  %s'%(SPACE_BUFFER, k, text))
    sb.append('')
    sb.append('  [paused]')
    print '\n'.join(sb)
    context.io.getchar()

@handle_keypress('p', ',')
def action_pickup(context):
    tile_container = context.player.ent.contained_by
    for idx, itm in enumerate(tile_container):
        if idx == 0:
            # this is the player
            pass
        elif idx == 1:
            # top item
            break
    if not itm.mobile:
        status(context, "Nothing to pick up")
    else:
        context.player.ent.inv.adopt(itm)
        status(context, "[%s] picked up"%itm.name)

@handle_keypress('P')
def action_drop(context):
    tile_container = context.player.ent.contained_by
    ent = inspect_mobile_items_in_container( context=context
                                           , container=context.player.ent.inv
                                           , heading="Select item to drop"
                                           , use_selection=True
                                           )
    if None == ent:
        return
    if None == ent.on_apply_fn:
        status(context, 'Nothing to do')
        return
    tile_container.adopt(ent)
    status(context, '[%s] dropped'%ent.name)

@handle_keypress(':')
def action_here(context):
    tile_container = context.player.ent.contained_by
    status(context, tile_container.things())
    pass

@handle_keypress(C_SW, 'z', 'b')
def action_go_sw(context):
    context.player.player_block().move(context.player.ent, 1)

@handle_keypress(C_SS, 'x', 'j')
def action_go_ss(context):
    context.player.player_block().move(context.player.ent, 2)

@handle_keypress(C_SE, 'c', 'n')
def action_go_se(context):
    context.player.player_block().move(context.player.ent, 3)

@handle_keypress(C_WW, 'a', 'h')
def action_go_ww(context):
    context.player.player_block().move(context.player.ent, 4)

@handle_keypress(C_EE, 'd', 'l')
def action_go_ee(context):
    context.player.player_block().move(context.player.ent, 6)

@handle_keypress(C_NW, 'q', 'y')
def action_go_nw(context):
    context.player.player_block().move(context.player.ent, 7)

@handle_keypress(C_NN, 'w', 'k')
def action_go_nn(context):
    context.player.player_block().move(context.player.ent, 8)

@handle_keypress(C_NE, 'e', 'u')
def action_go_ne(context):
    context.player.player_block().move(context.player.ent, 9)


# --------------------------------------------------------
#   fairies
# --------------------------------------------------------
#
# The program runs on fairies. A game 'turn' is comprised
# of all fairies working out what they want to do.
#
class PlayerFairy(object):
    def __init__(self, context, ent):
        self.context = context
        self.ent = ent
        self.health = 0
    def player_block(self):
        tile_container = self.ent.contained_by
        tile = tile_container.owner
        return tile.block
    def turn(self):
        "Have a turn"
        self.context.io.clear_screen()

        self.context.io.redraw( self.player_block().render()
                              , True
                              )
        self.context.status = []
        key = self.context.io.getchar()
        if key in KEYPRESS_ACTIONS:
            fn = KEYPRESS_ACTIONS[key]
            fn(self.context)
        else:
            self.context.status.append("Unknown command, %s"%key)
        #
        # Tell the player about the surface they're on
        self.context.status.append(str(self.context.clock))

def init_player_actor(context):
    middle = (ROWSIZE/2, COLSIZE/2)
    starting_tile = context.starting_block.d_tile[middle]

    ent = hatch_entity(context, 'player', starting_tile.container)
    context.player = PlayerFairy(context, ent)
    context.fairies.append(context.player)

class MotherNature(object):
    """Looks after the growth of plants and the like."""
    def __init__(self, context):
        self.context = context
        self.plants = []
    def turn(self):
        pass

def init_mother_nature(object):
    context.mother_nature = MotherNature(context)
    context.fairies.append(context.mother_nature)

def setup_fairies(context):
    init_player_actor(context)



# --------------------------------------------------------
#   alg
# --------------------------------------------------------
def setup_map(context):
    """Create rooms, and then tiles for walls and
    the enclosed areas of this block."""
    context.starting_block = create_block(context, ROWSIZE, COLSIZE)

    merge_block_onto_block( context=context
                          , source_block=create_house_block(context)
                          , target_block=context.starting_block
                          , t_ridx_offset=5
                          , t_cidx_offset=23
                          )

    merge_block_onto_block( context=context
                          , source_block=create_well_block(context)
                          , target_block=context.starting_block
                          , t_ridx_offset=7
                          , t_cidx_offset=42
                          )

    merge_block_onto_block( context=context
                          , source_block=create_outhouse_block(context)
                          , target_block=context.starting_block
                          , t_ridx_offset=14
                          , t_cidx_offset=1
                          )


# --------------------------------------------------------
#   util
# --------------------------------------------------------
class Timer(object):
    def __init__(self, context, turns, fn_action):
        self.context = context
        self.turns = turns
        self.fn_action = fn_action
    def tick(self):
        self.turns -= 1
        if self.turns == 0:
            fn_action()
            context.timers.remove(t)
def create_timer(context, turns, fn_action):
    t = Timer(context, 5, fn_action)
    context.timers.append(t)

def status(context, msg):
    context.status.append(msg)

def pause(context):
    print '[paused]'
    context.io.getchar()

def debug_pause(context, note):
    print note
    pause()


# --------------------------------------------------------
#   main
# --------------------------------------------------------
def main():
    context = Context()
    context.io = UnixIo(context)

    build_entity_stencils(context)
    build_craftbook(context)
    build_negotiators(context)

    setup_map(context)
    setup_fairies(context)

    status(context, "You've chosen to live in exile.")
    status(context, "This A-frame and the outhouses are your lot.")
    status(context, "It's unclear what is important in the world.")
    status(context, "[qweadzxc, s to act. Or use vi keys. ? shows help]")
    status(context, '')

    key = None
    while context.keep_running:
        context.clock += 1
        for actor in context.fairies:
            actor.turn()

    context.io.clear_screen()
    player_block = context.player.player_block()
    context.io.redraw( player_block.render()
                     , True
                     )

    if context.exit_message:
        print context.exit_message
        pause(context)
        pause(context)

